/*****************************************************************************/
/*                                                                           */
/*                                ICCONFIG.CC                                */
/*                                                                           */
/* (C) 1995-96  Ullrich von Bassewitz                                        */
/*              Wacholderweg 14                                              */
/*              D-70597 Stuttgart                                            */
/* EMail:       uz@ibb.schwaben.com                                          */
/*                                                                           */
/*****************************************************************************/



// $Id$
//
// $Log$
//
//



#include <string.h>

#include "check.h"
#include "str.h"

#include "icerror.h"
#include "icconfig.h"
#include "icdlog.h"



/*****************************************************************************/
/*                          struct IstecBaseConfig                           */
/*****************************************************************************/



unsigned char* IstecBaseConfig::Pack (unsigned char* Buf) const
// Pack the data of struct IstecBaseConfig into an array ready for
// transmission. The function returns Buf.
{
    unsigned char* B = Buf;

    *B++ = Connection;
    *B++ = AB_InterfaceCount;
    *B++ = Protocol;
    *B++ = VersionHigh;
    *B++ = VersionLow;
    *B++ = TFEAssignment;
    for (unsigned I = 0; I < bcMSNCount; I++) {
        memcpy (B, MSN [I], sizeof (MSN [I]));
        B += sizeof (MSN [I]);
    }
    *B++ = Music;
    *B++ = IntS0;
    *B++ = ExtS0;
    *B++ = QueryLoc1;
    *B++ = QueryLoc2;
    memcpy (B, Number1, sizeof (Number1));
    B += sizeof (Number1);
    memcpy (B, Number2, sizeof (Number2));
    B += sizeof (Number2);
    memcpy (B, MSNGroups, sizeof (MSNGroups));
    B += sizeof (MSNGroups);

    // Version 1.93 extensions follow.
    if ((unsigned (VersionHigh) * 100 + unsigned (VersionLow)) > 192) {

        // This is a new version
        *B++            = MusicPort;
        *B++            = (unsigned char) (CountryCode & 0x00FF);
        *B++            = (unsigned char) ((CountryCode >> 8) & 0x00FF);
        memcpy (B, TFELoc, sizeof (TFELoc));
        B += sizeof (TFELoc);

    } else {

        // This is an old version, clear out the additional fields
        *B++            = 0;            // MusicPort
        *B++            = 49;           // Germany, lo
        *B++            = 0;            //  ...     hi
        memset (B, 0, sizeof (TFELoc));
        B += sizeof (TFELoc);

    }

    // Clear out the reserved buffer space
    memset (B, 0, 6);       // Reserved space

    // Return the prepared buffer
    return Buf;
}



IstecBaseConfig& IstecBaseConfig::Unpack (const unsigned char* Buf)
// Unpack an array of char that contains data for an IstecBaseConfig struct.
// The function returns this.
{
    if ((Connection = *Buf++) > coMax) {
        WriteDebugLog (FormatStr ("IstecBaseConfig::Unpack: Got invalid value for connection: %02X",
                                  Connection));
        Connection = coMehrgeraete;
    }
    AB_InterfaceCount = *Buf++;
    if ((Protocol = *Buf++) > prMax) {
        WriteDebugLog (FormatStr ("IstecBaseConfig::Unpack: Got invalid value for protocol: %02X",
                                  Protocol));
        Protocol = prDSS1;
    }
    VersionHigh   = *Buf++;
    VersionLow    = *Buf++;
    TFEAssignment = *Buf++;
    if ((TFEAssignment != 0) &&
        (TFEAssignment < 21 || TFEAssignment >= 21 + AB_InterfaceCount)) {
        WriteDebugLog (FormatStr ("IstecBaseConfig::Unpack: Got invalid value for TFE assignment: %02X",
                                  TFEAssignment));
        TFEAssignment = 21;
    }
    for (unsigned I = 0; I < bcMSNCount; I++) {
        memcpy (MSN [I], Buf, sizeof (MSN [I]));
        Buf += sizeof (MSN [I]);
    }
    Music             = (*Buf++ != 0);
    IntS0             = *Buf++;
    ExtS0             = *Buf++;
    QueryLoc1         = *Buf++;
    QueryLoc2         = *Buf++;
    memcpy (Number1, Buf, sizeof (Number1));
    Buf += sizeof (Number1);
    memcpy (Number2, Buf, sizeof (Number2));
    Buf += sizeof (Number2);
    memcpy (MSNGroups, Buf, sizeof (MSNGroups));
    Buf += sizeof (MSNGroups);

    // Version 1.93 extensions follow
    if ((unsigned (VersionHigh) * 100 + unsigned (VersionLow)) > 192) {

        // This is a new version
        MusicPort   = *Buf++;
        CountryCode = unsigned (Buf [1]) * 0x100 + unsigned (Buf [0]);
        Buf += 2;
        memcpy (TFELoc, Buf, sizeof (TFELoc));
        Buf += sizeof (TFELoc);

    } else {

        // This is an old version, clear out the additional fields
        MusicPort   = 0;            // None
        CountryCode = 49;           // Germany
        memset (TFELoc, 0, sizeof (TFELoc));

    }

    // Return the prepared struct
    return *this;
}



IstecBaseConfig& IstecBaseConfig::Unpack (const IstecMsg& Msg)
// Unpack an array of char that contains data for an IstecBaseConfig struct.
// The function returns this.
{
    PRECONDITION (Msg.Size >= 94 && Msg.At (0) == 0x17);
    return Unpack (&Msg.Data [1]);
}



int operator == (const IstecBaseConfig& lhs, const IstecBaseConfig& rhs)
{
    // For simplicity, convert to an array, then do a memcmp
    unsigned char Left [BaseConfigSize], Right [BaseConfigSize];
    return memcmp (lhs.Pack (Left), rhs.Pack (Right), BaseConfigSize) == 0;
}



int operator != (const IstecBaseConfig& lhs, const IstecBaseConfig& rhs)
{
    // For simplicity, convert to an array, then do a memcmp
    unsigned char Left [BaseConfigSize], Right [BaseConfigSize];
    return memcmp (lhs.Pack (Left), rhs.Pack (Right), BaseConfigSize) != 0;
}



/*****************************************************************************/
/*                              struct DevConfig                             */
/*****************************************************************************/



void DevConfig::ClearExtNum ()
// Clear the external number
{
    memset (ExtNum, 0xFF, sizeof (ExtNum));
}



unsigned char* DevConfig::Pack (unsigned char* Buf, double FirmwareVersion) const
// Pack the data of struct DevConfig into an array ready for transmission.
// The function returns Buf.
{
    unsigned char* B = Buf;

    *B++ = DevNum;
    *B++ = DialCaps;
    *B++ = Service;
    *B++ = Reroute;
    *B++ = ChargePulse;
    *B++ = PIN [0];
    *B++ = PIN [1];
    memcpy (B, ExtNum, sizeof (ExtNum));
    B += sizeof (ExtNum);

    // Version 1.93+1.94 extensions follow
    if (FirmwareVersion > 1.92) {

        // This is a new version config record
        *B++ = TerminalMode;

        if (FirmwareVersion > 1.93) {
            *B++ = InternalKnock;
            *B++ = ExternalKnock & 0xFF;
            *B++ = (ExternalKnock >> 8) & 0xFF;
            *B++ = 0;           // Undocumented byte 23
        } else {
            // Clear out the additional fields
            memset (B, 0, 4);
        }

    } else {

        // Clear out the additional fields
        memset (B, 0, 5);

    }

    // Return the prepared buffer
    return Buf;
}



DevConfig& DevConfig::Unpack (const unsigned char* Buf, double FirmwareVersion)
// Unpack an array of char that contains data for a DevConfig struct.
// The function returns *this and corrects invalid raw values.
{
    unsigned svMax = FirmwareVersion < 1.92? svAnrufbeantworter : svKombi;

    DevNum = *Buf++;
    if ((DialCaps = *Buf++) > dcMax) {
        WriteDebugLog (FormatStr ("IstecDevConfig::Unpack:: Got invalid value for DialCaps: %02X",
                                  DialCaps));
        DialCaps = dcKeine;
    }
    if ((Service = *Buf++) > svMax) {
        WriteDebugLog (FormatStr ("IstecDevConfig::Unpack:: Got invalid value for Service: %02X",
                                  Service));
        Service = svFernsprechen;
    }
    Reroute         = *Buf++ != 0;
    ChargePulse     = (*Buf++ != 0);
    PIN [0]         = *Buf++;
    PIN [1]         = *Buf++;
    memcpy (ExtNum, Buf, sizeof (ExtNum));
    Buf += sizeof (ExtNum);

    // Version 1.93+1.94 extensions follow
    if (FirmwareVersion > 1.92) {

        // This is a new version config record
        TerminalMode = (*Buf++ != 0);

        if (FirmwareVersion > 1.93) {
            InternalKnock = *Buf++;
            ExternalKnock = Buf [0] + 256 * Buf [1];
            Buf += 2;
        } else {
            // Clear out the additional fields
            InternalKnock = 0;
            ExternalKnock = 0;
        }

    } else {

        // Clear out the additional fields
        TerminalMode  = 0;
        InternalKnock = 0;
        ExternalKnock = 0;

    }

    // Return the prepared struct
    return *this;
}



DevConfig& DevConfig::Unpack (const IstecMsg& Msg, double aFirmware)
// Unpack an array of char that contains data for a DevConfig struct.
// The function returns *this and corrects invalid raw values.
{
    PRECONDITION (Msg.Size >= 18 && Msg.At (0) == 0x16);
    return Unpack (&Msg.Data [1], aFirmware);
}



int operator == (const DevConfig& lhs, const DevConfig& rhs)
{
    // For simplicity, convert to an array, then do a memcmp
    unsigned char Left [DevConfigSize], Right [DevConfigSize];
    return memcmp (lhs.Pack (Left), rhs.Pack (Right), DevConfigSize) == 0;
}



int operator != (const DevConfig& lhs, const DevConfig& rhs)
{
    // For simplicity, convert to an array, then do a memcmp
    unsigned char Left [DevConfigSize], Right [DevConfigSize];
    return memcmp (lhs.Pack (Left), rhs.Pack (Right), DevConfigSize) != 0;
}



/*****************************************************************************/
/*                             class IstecCharges                            */
/*****************************************************************************/



void IstecCharges::Clear ()
// Clear the charges
{
    for (unsigned I = 0; I < IstecDevCount; I++) {
        Charges [I] = 0;
    }
}



unsigned& IstecCharges::operator [] (unsigned Device)
// Return a reference to the charges of a specific device
{
    PRECONDITION (Device < IstecDevCount);
    return Charges [Device];
}



const unsigned& IstecCharges::operator [] (unsigned Device) const
// Return a reference to the charges of a specific device
{
    PRECONDITION (Device < IstecDevCount);
    return Charges [Device];
}



unsigned char* IstecCharges::Pack (unsigned char* Buf) const
// Pack the data of IstecCharges into an array ready for transmission.
// The function returns Buf.
{
    unsigned char* B = Buf;

    for (unsigned I = 0; I < IstecDevCount; I++) {

        // Check range
        CHECK (Charges [I] <= 0xFFFF);

        // Get the value
        u16 Val = (u16) Charges [I];

        // Store the value into the array in little endian format
        *B++ = (unsigned char) Val;
        *B++ = (unsigned char) (Val / 256);

    }

    // Return the prepared buffer
    return Buf;
}



IstecCharges& IstecCharges::Unpack (const unsigned char* Buf)
// Unpack an array of char that contains data for an IstecCharges struct.
// The function returns this.
{
    for (unsigned I = 0; I < IstecDevCount; I++) {

        // Convert the value from little endian format
        Charges [I] = Buf [0] + (unsigned)Buf [1] * 256;

        // Next one
        Buf += 2;
    }

    // Return the prepared struct
    return *this;
}



int operator == (const IstecCharges& lhs, const IstecCharges& rhs)
{
    for (unsigned I = 0; I < IstecDevCount; I++) {
        if (lhs.Charges [I] != rhs.Charges [I]) {
            return 0;
        }
    }
    return 1;
}



int operator != (const IstecCharges& lhs, const IstecCharges& rhs)
{
    for (unsigned I = 0; I < IstecDevCount; I++) {
        if (lhs.Charges [I] == rhs.Charges [I]) {
            return 0;
        }
    }
    return 1;
}



/*****************************************************************************/
/*                                   Code                                    */
/*****************************************************************************/



unsigned char* ToBCD (const char* S, unsigned char* T, unsigned TSize)
// Convert the number string in S to a BCD representation in T, filling unused
// digits with 'F'. The function returns T.
{
    // Check the parameters
    PRECONDITION (S != NULL && T != NULL && TSize > 0);

    // Calculate the maximum digit count and make shure, T is big enough
    unsigned Count = TSize * 2;
    unsigned Len   = strlen (S);
    if (Len > Count) {
        // OOPS - S is too long, what now?
        IstecError (msErrorIntIgnored);
        Len = Count;
    }

    // Pre-fill the target string with FF
    memset (T, 0xFF, TSize);

    // Convert to BCD and store
    int LowNibble = 1;
    unsigned char* P = T;
    while (Len--) {

        if (LowNibble) {
            *P = (*S - '0') | 0xF0;
            LowNibble = 0;
        } else {
            *P = ((*S - '0') << 4) | (*P & 0x0F);
            P++;
            LowNibble = 1;
        }
        S++;
    }

    // Return the target buffer
    return T;
}



char* FromBCD (char* S, const unsigned char* T, unsigned TSize)
// Convert the BCD string in T into an ASCII representation in S. Conversion
// stops if an invalid BCD char or TSize is reached. A trailing zero is added
// to S. It is assumed that S is big anough to hold the resulting string (S
// must be TSize*2+1 chars in size).
// The function returns S.
{
    // Check the parameters
    PRECONDITION (S != NULL && T != NULL && TSize > 0);

    // Calculate the nibble count
    TSize *= 2;

    // Conversion loop
    char* Q = S;
    unsigned LowNibble = 1;
    while (TSize--) {

        int C;
        if (LowNibble) {
            C = *T & 0x0F;
            LowNibble = 0;
        } else {
            C = (*T & 0xF0) >> 4;
            T++;
            LowNibble = 1;
        }

        if (C < 10) {
            *Q++ = '0' + C;
        } else {
            // Invalid char, end reached
            break;
        }
    }

    // Add a trailing zero
    *Q = '\0';

    // return the result
    return S;
}



unsigned char* ToPascal (const char* S, unsigned char* T, unsigned TSize)
// Convert the C style string in S to the pascal string in T and return T.
{
    // Check parameters
    PRECONDITION (S != NULL && T != NULL && TSize > 0);

    // Copy the string
    unsigned char* Q = T + 1;
    unsigned Length = 0;
    while (*S && TSize--) {
        *Q++ = *S++;
        Length++;
    }

    // Pad the target string with zeros
    while (TSize--) {
        *Q++ = '\0';
    }

    // Set the string length
    *T = Length;

    // Return the result
    return T;
}



char* FromPascal (char* S, const unsigned char* T, unsigned SSize)
// Convert the pascal style string in T into a C like string in S and return S.
{
    // Check parameters
    PRECONDITION (S != NULL && T != NULL && SSize > 0);

    // Get the string length and restrict it if S is not big enough
    unsigned Length = *T++;
    if (Length > SSize-1) {
        Length = SSize-1;
    }

    // Copy the string
    char* Q = S;
    while (Length--) {
        *Q++ = *T++;
    }

    // Add a trailing zero
    *Q = '\0';

    // Return the result
    return S;
}



