/*****************************************************************************/
/*                                                                           */
/*                               ISTECMSG.CC                                 */
/*                                                                           */
/* (C) 1996     Ullrich von Bassewitz                                        */
/*              Wacholderweg 14                                              */
/*              D-70597 Stuttgart                                            */
/* EMail:       uz@ibb.schwaben.com                                          */
/*                                                                           */
/*****************************************************************************/



// $Id$
//
// $Log$
//
//



// Defines a class that holds an istec message



#include "check.h"

#include "icver.h"
#include "icdlog.h"
#include "icdiag.h"
#include "istecmsg.h"



/*****************************************************************************/
/*                              class IstecMsg                               */
/*****************************************************************************/



IstecMsg::IstecMsg (unsigned MsgSize, const unsigned char* Buf):
    ErrCode (0),
    Size (MsgSize)
// Create a message
{
    if (MsgSize > 0) {
        Data = new unsigned char [MsgSize];
        if (Buf) {
            // We got data, copy it
            memmove (Data, Buf, MsgSize);
        }
    } else {
        Data = 0;
    }
}



IstecMsg::~IstecMsg ()
{
    delete [] Data;
}



String IstecMsg::AsciiData ()
{
    return ::AsciiData (Data, Size);
}



int IstecMsg::IsDiagMsg ()
// Return true if the message is a diagnostic message
{
    // Unnecessary override for gcc
    return ::IsDiagMsg (Data, Size);
}



int IstecMsg::IsChargeInfo ()
// Return true if the message is a charge info message
{
    // Unnecessary override for gcc
    return ::IsChargeInfo (Data, Size);
}



unsigned char& IstecMsg::At (unsigned Index)
// Access the Data member bounds checked.
{
    PRECONDITION (Index < Size);
    return Data [Index];
}



const unsigned char& IstecMsg::At (unsigned Index) const
// Access the Data member bounds checked.
{
    PRECONDITION (Index < Size);
    return Data [Index];
}



/*****************************************************************************/
/*                     Other Istec Message related stuff                     */
/*****************************************************************************/



String IstecMsgDesc (const unsigned char* Data)
// Return a string describing the Istec message. This is for debug purposes
// only, the string is hardcoded, not loaded from the resource.
{
    PRECONDITION (Data != 0);

    switch (Data [0]) {

        case 0x02:
            return "Are you there?";

        case 0x05:
            return "Write charge data";

        case 0x06:
            return "Request charge data";

        case 0x07:
            return "Write device configuration";

        case 0x08:
            return "Request device configuration";

        case 0x09:
            return "Write base configuration";

        case 0x0a:
            return "Request base configuration";

        case 0x0c:
            return "End configuration";

        case 0x11:
            return "End configuration (ACK)";

        case 0x12:
            return "Hi! Nice to meet you!";

        case 0x13:
            return "Error reply (NACK)";

        case 0x15:
            return "Charge data";

        case 0x16:
            return "Device configuration data";

        case 0x17:
            return "Base configuration data";

        case 0x18:
            return "Write device configuration acknowledged";

        case 0xdd:
            return "Diagnostic message";

        default:
            return "Unknown message";

    }
}



String AsciiData (unsigned char* Data, unsigned Size)
// Create an ascii string with the hex representation of the message data
{
    // Check the programmer :-)
    PRECONDITION (Data != NULL && Size > 0);

    String S (Size * 3);

    // Create the string
    for (unsigned I = 0; I < Size; I++) {
        S += FormatStr ("%02X ", Data [I]);
    }

    // Add a short description of the message if VerboseDebugLog is enabled
    if (IsDiagMsg (Data, Size)) {
        S += DiagMsgDesc (Data);
    } else {
        S += IstecMsgDesc (Data);
    }

    // Return the result
    return S;
}



int IsDiagMsg (unsigned char* Data, unsigned Size)
// Return true if the given message is a diagnostic message
{
    PRECONDITION (Data != NULL);
    return Size == 6 && Data [0] == 0xdd;
}



int IsChargeInfo (unsigned char* Data, unsigned Size)
// Return true if the message contains the ISTEC charges
{
    return Size == 129 && Data [0] == 0x15;
}



unsigned SendMsgSize (const unsigned char* Data, unsigned BufSize)
// The function returns the real size of the given message to send. Since
// the messages given to this function are created by the programmer (that
// is me - hi!), a non determinable message size is a programming error.
// Note: BufSize is the size of the message buffer, the real message that
// is sent maybe smaller than this value.
{
    unsigned Version;

    // Check the given parameter
    PRECONDITION (Data != NULL && BufSize > 0);

    switch (Data [0]) {

        case 0x02:
        case 0x06:
        case 0x08:
        case 0x0a:
        case 0x0c:
            return 1;

        case 0x05:
            return 129;

        case 0x07:
            return (FirmwareVersion < 1.93) ? 18 : 22;

        case 0x09:
            // This is ugly, but I cannot help: Use the firmware version
            // from the message instead of the saved one, since we may
            // not know about the version when receiving this message.
            // To be shure not to access non-existing message members,
            // check the length first.
            CHECK (BufSize >= 94);
            Version = unsigned (Data [4]) * 100 + unsigned (Data [5]);
            return (Version < 193) ? 94 : 107;

        case 0xdd:
            return 6;

        default:
            FAIL ("SendMsgSize: Trying to create invalid message!");
            return 0;
    }
}



unsigned RecMsgSize (unsigned char* Data, unsigned Size)
// Calculate the size of a message. First byte is opcode.
{
    // Check the given parameter
    PRECONDITION (Data != NULL && Size > 0);

    switch (Data [0]) {

        case 0x11:
        case 0x12:
            return 1;

        case 0x13:
            WriteDebugLog ("Got error reply: " + AsciiData (Data, Size));
            return 2;

        case 0x15:
            return 129;

        case 0x16:
            if (FirmwareVersion < 1.93) {
                return 18;
            } else if (FirmwareVersion < 1.95) {
                return 22;
            } else {
                return 23;
            }

        case 0x17:
            // This is ugly, but I cannot help: Use the firmware version
            // from the message instead of the saved one, since this we may
            // not know about the version when receiving this message.
            // To be shure not to access non-existing message members,
            // check the length first.
            if (Size != 94 && Size != 107) {
                WriteDebugLog ("Error: Invalid message length: " +
                               AsciiData (Data, Size));
                return Size;
            } else {
                // Get the version
                unsigned Version = unsigned (Data [4]) * 100 + unsigned (Data [5]);
                return Version < 193 ? 94 : 107;
            }

        case 0x18:
            return 2;

        case 0xdd:
            return 6;

        default:
            WriteDebugLog ("Warning: Cannot determine message length: " +
                           AsciiData (Data, Size));
            return Size;
    }
}



